import numpy as np
import wavio
import matplotlib.pyplot as plt
from scipy.signal.windows import hann
# To Do
# - implement filter Q-parameter mapping

class Stream:
    """ Stream object representing audio samples"""
    def __init__(self, length, samprate=44100, ltype='seconds'):

        # variables we want to keep constant
        self.samprate = samprate
        self._nyqfrq = 0.5*self.samprate
        
        if ltype == 'seconds':
            self.length = length
            self._nsamp_stream = int(samprate * length)
        elif ltype == 'samples':
            self._nsamp_stream = length
            self.length = length / samprate
            
            
        # sample values initialised to 0 (silence)
        self.values =  np.zeros(self._nsamp_stream)

        # private stream for keeping track of buffered stream
        self._bvalues =  np.zeros(self._nsamp_stream)
        
        # sample numbers for indexing
        self.samples = np.arange(self._nsamp_stream, dtype=int)

        # time at which each sample occurs
        self.samptime = self.samples / self.samprate
        
    def bufferize(self, bufflength=0.1):
        """ wrapper to initialise Buffers subclass """
        self.buffers = Buffers(self, bufflength)

    def consolidate_buffers(self):
        """ wrapper to reassign stream values to consolidated stream """
        self.values = self.buffers.to_stream()
        
    def filt_sweep(self, ffunc, fmap, qmap=lambda x:x*0 + 0.1,
                   flo=20, fhi=2.205e4, qlo=0.5, qhi=10):
        """
        ffunc: function that applies filter 
        fmap: mapping function representing filter cutoff sweep
        qmap: mapping function for a filters Q parameter, default: lambda:None
        flo: lowest frequency of sweep in Hz, default 20
        fhi: highest frequency of sweep in Hz, default 22.05 kHz
        """
        if not hasattr(self, "buffers"):
            Exception("needs bufferized stream, please run 'bufferize' method first.")
        buffers = self.buffers

        # array to regularly sample maps at each buffer
        x = np.linspace(0, 1, buffers._nbuffs_tot)

        lfhi = np.log10(fhi)
        lflo = np.log10(flo)
        
        # obtain buffer values from maps, with cutoff sweep in units
        # of the nyquist frequency
        svals = pow(10., fmap(x)*(lfhi-lflo)+lflo)/self._nyqfrq
        qvals = (qmap(x)*(qhi-qlo)+qlo)

        # loop over buffers, applying appropriate filtering to each 
        for i in range(buffers._nbuffs):
            i2 = 2*i
            buffers.buffs_tile[i] = ffunc(buffers.buffs_tile[i], svals[i2], qvals[i2])
        for i in range(buffers._nbuffs-1):
            i2 = 2*i+1
            buffers.buffs_olap[i] = ffunc(buffers.buffs_olap[i], svals[i2], qvals[i2])

        # finally, consolidate buffers to apply effect to stream
        self.consolidate_buffers()

    def get_sampfracs(self):
        self.sampfracs = np.linspace(0, 1, self.values.size)
        
    def save_wav(self, filename):
        """ save audio stream to wav file, specified by filename"""
        wavio.write(filename, self.values, self.samprate, sampwidth=3) 
        
    def reset(self):
        """ zero audio stream and buffers if present """
        self.values *= 0.
        if hasattr(self, "buffers"):
            self.buffers.buffs_tile *= 0.
            self.buffers.buffs_olap *= 0.        
        
class Buffers: 
    def __init__(self, stream, bufflength=0.1):
        nbuff = stream.samprate*bufflength
        if nbuff < 20:
            Exception(f"Error: buffer length {nbuff} samples below "
                      "lower limit of 20, with specified bufflength "
                      "{bufflength} seconds and sample rate {self.samprate} Hz")

        # force buffer length to an even number of samples
        self._nsamp_halfbuff = int(stream.samprate*bufflength) // 2
        self._nsamp_buff = 2 * self._nsamp_halfbuff

        # minimum number of tiled buffers to completely enclose stream 
        self._nbuffs = 1+(stream._nsamp_stream//self._nsamp_buff)

        # total number buffers including overlaps
        self._nbuffs_tot = 2*self._nbuffs-1

        # tent function for linearly x-fading buffers on recombination
        # self.fade = 1.-abs(np.linspace(1,-1, self._nsamp_buff))
        self.fade = hann(self._nsamp_buff)

        # pad the stream up to an exact multiple of buffer sample length
        self.nsamp_padstream = self._nbuffs * self._nsamp_buff
        self.nsamp_pad = self.nsamp_padstream-stream._nsamp_stream
        self.olap_pad = self.nsamp_pad-self._nsamp_halfbuff
        self.olap_lim = min(stream._nsamp_stream, stream._nsamp_stream+self.olap_pad)
        
        # construct tile and overlap buffer arrays
        self.buffs_tile = np.pad(stream.values, (0,self.nsamp_pad)
                                 ).reshape((self._nbuffs, self._nsamp_buff))
        self.buffs_olap = np.pad(stream.values[self._nsamp_halfbuff:self.olap_lim],
                                 (0,max(0, self.olap_pad))
                                 ).reshape((self._nbuffs-1), self._nsamp_buff)

    def to_stream(self):
        """ reconstruct stream by x-fading buffers """
        # apply fades to buffers, first special edge cases...
        self.buffs_tile[0,self._nsamp_halfbuff:] *= self.fade[self._nsamp_halfbuff:]
        self.buffs_tile[-1,:self._nsamp_halfbuff] *= self.fade[:self._nsamp_halfbuff]

        # ...then to remaining buffers
        self.buffs_tile[1:-1,:] *= self.fade.T
        self.buffs_olap *= self.fade.T

        # reconstruct stream
        padded_stream = np.zeros(self.nsamp_padstream)
        padded_stream += self.buffs_tile.flatten()
        flat_olaps    = self.buffs_olap.flatten()
        padded_stream[self._nsamp_halfbuff:-self._nsamp_halfbuff] += flat_olaps

        # remove padding on returning reconstructed stream
        return padded_stream[:-self.nsamp_pad]



from functools import reduce
import operator
import numpy as np
from scipy.interpolate import interp1d
from contextlib import contextmanager,redirect_stderr,redirect_stdout
from os import devnull
from io import StringIO 
import sys
from pathlib import Path

# Some utility classes (these may graduate to somewhere else eventually)

class NoSoundDevice:
    """
    drop-in replacement for sounddevice module if not working,
    so can still use other functionality.
    """
    def __init__(self, err):
        self.err = err
    def play(self, audio, rate, blocking=1):
        raise self.err

class Equaliser:
    def __init__(self):

        self.factor_rms = None
        parpath = Path(f"{Path(__file__).parent}","data","params.csv")
        # Read in parameters for ISO 226:2024 standard.
        pars = np.genfromtxt(parpath, delimiter=',', names=True)
        self.parfuncs = {}
        for c in pars.dtype.names:
            if c == 'freq':
                continue
            self.parfuncs[c] = interp1d(np.log10(pars['freq']),
                                        pars[c],
                                        fill_value='extrapolate')
            
    def get_relative_loudness_norm(self, freq, phon=70.):
        """
        Relative normalisation of sound frequencies
        To compensate for pereceptual loudness, following
        the ISO 226:2024 standard.

        Args:
          freq (:obj:`array-like`) audio frequencies in Hz
          phon (:obj:`float`) listening level for a 1 kHz
            note 

        Returns:
          rnorm (:obj:`array-like`) volume normalisation
            for spectra
        """
        lfreq = np.log10(freq)
        L_U = self.parfuncs['L_U'](lfreq)
        alpha_f = self.parfuncs['alpha_f'](lfreq)
        T_f = self.parfuncs['T_f'](lfreq)

        A = pow(4e-10, 0.3 - alpha_f)
        B = pow(10., (phon)*3e-2) - pow(10, 7.2e-2)
        C = pow(10., alpha_f * 0.1*(T_f + L_U))

        L_f = 10*np.log10(A*B + C)/alpha_f - L_U
        norm = pow(10., (L_f - phon)/20)
        rnorm = norm/norm.max()
        return rnorm

    
# a load of utility functions used by STRAUSS

def nested_dict_reassign(fromdict, todict):
    """recurse through dictionaries and sub-dictionaries"""
    for k, v in fromdict.items():
        if isinstance(v, dict):
            # recurse through nested dictionaries
            nested_dict_reassign(v, todict[k])
        else:
            # reassign todict value
            todict[k] = v

def nested_dict_fill(fromdict, todict):
    """recurse through dictionaries and sub-dictionaries"""
    for k, v in fromdict.items():
        if k not in todict:
            # assign todict value
            todict[k] = v
        elif isinstance(v, dict):
            # recurse through nested dictionaries
            nested_dict_fill(todict[k], v)
            
def nested_dict_idx_reassign(fromdict, todict, idx):
    """recurse through dictionaries and sub-dictionaries"""
    for k, v in fromdict.items():
        if isinstance(v, dict):
            # recurse through nested dictionaries
            nested_dict_idx_reassign(todict[k], v, idx)
        else:
            # reassign todict value
            todict[k] = v[idx]

def reassign_nested_item_from_keypath(dictionary, keypath, value):
    """
    dictionary: dict, dict object to reassign values of
    keypath: str, 'a/b/c' corresponds to dict['a']['b']['c'] 
             or (for Windows systems): str, 'a\b\c' corresponds to dict['a']['b']['c'] 
    value: any, value to reassign dictionary value with
    """
    p = Path(keypath)
    keylist = list(p.parts)
    get_item = lambda d, kl: reduce(operator.getitem, kl, d)
    get_item(dictionary, keylist[:-1])[keylist[-1]] = value
            
def linear_to_nested_dict_reassign(fromdict, todict):
    """iterate through a linear dictionary to reassign nested values
    using keypaths (d1['a/b/c'] -> d2['a']['b']['c'], d1['a']->d2['a'])"""
    for k, v in fromdict.items():
        reassign_nested_item_from_keypath(todict, k, v)
            
def const_or_evo_func(x):
    """if x is callable, return x, else provide a function that returns x"""
    if callable(x):
        return x
    else:
        return lambda y: y*0 + x

def const_or_evo(x,t):
    """if x is callable, return x(t), else return x"""
    if callable(x):
        return x(t)
    else:
        return x

def rescale_values(x, oldlims, newlims):
    """ rescale x values to range limits such that 0-1 is mapped to limits[0]-limits[1] """
    olo, ohi = oldlims
    nlo, nhi = newlims
    descale = np.clip((x - olo) / (ohi-olo), 0 , 1)
    return (nhi-nlo)*descale + nlo
    
def resample(rate_in, samprate, wavobj):
    """ resample audio from original samplerate to required samplerate """
    duration = wavobj.shape[0] / rate_in

    time_old  = np.linspace(0, duration, wavobj.shape[0])
    time_new  = np.linspace(0, duration,
                            int(wavobj.shape[0] * samprate / rate_in))

    interpolator = interp1d(time_old, wavobj.T)
    new_wavobj = np.round(interpolator(time_new).T).astype(wavobj.dtype)
    return(new_wavobj)

@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, 'w') as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)
            
class Capturing(list):
    """ Context manager for handling stdout (see https://stackoverflow.com/a/16571630) """
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout



import numpy as np
import pychord as chrd
import re

# Tune note system to A440 standard
# Equal temperement

tuneC0 = 440 * pow(2, (-9./12)-4) 

notecount = np.arange(12)
notesharps  = ["C","C#","D","D#","E","F","F#","G","G#","A","A#","B"]
noteflats = ["C","Db","D","Eb","E","F","Gb","G","Ab","A","Bb","B"]
semitone_dict = {**dict(zip(notesharps, notecount)),
                 **dict(zip(noteflats, notecount))}

def parse_note(notename):
    """ 
    Takes scientific pitch name and returns frequency in Hz.
    flat and sharp numbers supported
    """
    nsplit = re.findall("(\D+|\d+)", notename)
    semi = semitone_dict[nsplit[0]]/12.
    octv  = int(nsplit[1])
    return tuneC0*pow(2.,semi+octv)

def parse_chord(chordname, rootoct=3):
    chord = chrd.Chord(chordname)
    notes = chord.components_with_pitch(rootoct)
    frqs = []
    for n in notes:
        frqs.append(parse_note(n))
    return np.array(frqs)

def chord_notes(chordname, rootoct=3):
    chord = chrd.Chord(chordname)
    notes = chord.components_with_pitch(int(rootoct))
    return notes

def mkey_to_note(val):
    from strauss.notes import notesharps
    octv = val // 12 - 1
    semi = val % 12
    return f'{notesharps[semi]}{octv}'



import numpy as np
import scipy.signal as sig

def LPF1(data, cutoff, q, order=5):
    b, a = sig.butter(order, cutoff, btype='low', analog=False)
    y = sig.lfilter(b, a, data)
    return y

def HPF1(data, cutoff, q, order=5):
    b, a = sig.butter(order, cutoff, btype='high', analog=False)
    y = sig.lfilter(b, a, data)
    return y




""" :obj:`sonification`: generate sonification, combining submodules.

This Submodule handles the combining of all the constituent
subroutines into  a single :obj:`sonification` object that can then
render and output/save the resultant sonification. This handles
feeding of information between :obj:`strauss` modules, including
taking the :obj:`sources` mapping, applying any musical constraints
from :obj:`score` running the :obj:`generators` to make sound and
combining them into the output channels for the overall spatialised
sonificiation.

Todo:
  * Delegate more musical process to the :obj:`score` module
"""

from .stream import Stream
from .channels import audio_channels
from .utilities import const_or_evo, nested_dict_idx_reassign, NoSoundDevice
from .tts_caption import render_caption
import numpy as np
import matplotlib.pyplot as plt
import sys
import os
import ffmpeg as ff
import wavio as wav
import IPython.display as ipd
from IPython.core.display import display
from scipy.io import wavfile
import warnings
import tempfile
from pathlib import Path
try:
    import sounddevice as sd
except (OSError, ModuleNotFoundError) as sderr:
    sd = NoSoundDevice(sderr)
try:
    from tqdm import tqdm
except ModuleNotFoundError:
    tqdm = list

class Sonification:
    """Representing the overall sonification

    This class combines the data sources, musical score constraints
    and generator together to generate and render the ultimate
    sonification for saving or playing in the :obj:`jupyter-notebook`
    environment 

    Args:
      score (:class:`~strauss.score.Score`): Sonification :obj:`Score`
    	object 
      sources (:class:`~strauss.sources.Source`): Sonification
    	:obj:`Sources` child object (:class:`~strauss.sources.Events`
    	or :class:`~strauss.sources.Objects`)  
      generator (:class:`~strauss.generator.Generator`): Sonification
    	:obj:`Generator` child object
    	(:class:`~strauss.generator.Synthesizer` or
    	:class:`~strauss.generator.Sampler`)
      audio_setup (:obj:`str`) The requested audio setup preset to
    	pass to :class:`~strauss.channels.audio_channels`
      samprate (:obj:`int`) Integer sample rate in samples per second
        (Hz), typically :obj:`44100` or :obj:`48000` for most audio
    	applications
      ttsmodel (:obj:`str`) The text-to-speech model used for captions. 

    Todo:
      * Support custom audio setups here too.
    """
    def __init__(self, score, sources, generator, audio_setup='stereo',
                 caption=None, samprate=48000,
                 ttsmodel=Path('tts_models','en','jenny', 'jenny')):

        # sampling rate in Hz
        self.samprate = samprate
        
        # tts model name
        self.ttsmodel = str(ttsmodel)
        
        # caption
        self.caption = caption
        
        # sonification owns an instance of the Score
        self.score = score
        
        # sonification owns an instance of the Sources
        self.sources = sources

        # sonification owns an instance of the Generator
        self.generator = generator
        
        # set up the audio channel routing for the sonification
        self.channels = audio_channels(setup=audio_setup)

        # check Generator and Sonification sampling rates match...
        if self.samprate != self.generator.samprate:
            # if not, revert to Generator sampling rate.
            warnings.warn("warning: global and generator sampling rates disagree, " \
            f"reverting to generator value of {self.generator.samprate} Hz")
            self.samprate = self.generator.samprate
        
        # ...and the corresponding Stream objects 
        self.out_channels = {}
        for c in range(self.channels.Nmics):
            self.out_channels[str(c)] = Stream(self.score.length, self.samprate)

    def render(self, downsamp=1):
        """Render the sonification.
        
        Generates the sonification by running the  Synthesizer
        :func:`~strauss.generator.Synthesizer.play` or Sampler
        :func:`~strauss.generator.Sampler.play` functions, and
        combining these into the output channel streams using any
        spatialisation for the specified
        :class:`~strauss.channels.audio_channels`. 

        Args:
          downsamp (optional, :obj:`int`): Optionally downsample
          sources for multi-source sonifications for a quicker test
          render by some integer factor.
        """

        # first determine if time is provided, if not assume all start at zero
        # and last the duration of sonification

        if "time" not in self.sources.mapping:
            self.sources.mapping['time'] = [0.] * self.sources.n_sources
            self.sources.mapping['note_length'] = [self.score.length] * self.sources.n_sources
            
        # index each chord
        cbin = np.digitize(self.sources.mapping['time'], self.score.fracbins, 0)
        cbin = np.clip(cbin-1, 0, self.score.nchords-1)

        # pitch rank of each source divided by the number of sources
        pitchfrac = np.empty_like(self.sources.mapping['pitch'])
        if self.score.pitch_binning == 'adaptive':
            pitchfrac[np.argsort(self.sources.mapping['pitch'])] = np.arange(self.sources.n_sources)/self.sources.n_sources
        elif self.score.pitch_binning == 'uniform':
            pitchfrac = np.clip(self.sources.mapping['pitch'], 0, 9.999999e-1)
            
        # get some relevant numbers before iterating through sources
        Nsamp = self.out_channels['0'].values.size
        lastsamp = Nsamp - 1
        Nchan = len(self.out_channels.keys())
        indices = range(0,self.sources.n_sources, downsamp)

        for source in tqdm(indices):

            # index note properties
            t = self.sources.mapping['time'][source]
            tsamp = int(Nsamp * t)
            chord = self.score.note_sequence[cbin[source]]
            nints = self.score.nintervals[cbin[source]]
            pitch = pitchfrac[source]
            note = chord[int(pitch * nints)]

            # make dictionary for feeding to play function with each notes properties
            sourcemap = {}
            # for k in self.sources.mapping.keys():
            #     sourcemap[k] = self.soures.mapping[k][source]
            nested_dict_idx_reassign(self.sources.mapping, sourcemap, source)

            sourcemap['note'] = note

            # run generator to play each note
            sstream = self.generator.play(sourcemap)
            playlen = sstream.values.size
            if 'phi' in sourcemap:
                azi     = const_or_evo(sourcemap['phi'], sstream.sampfracs) * 2 * np.pi
            elif 'azimuth' in sourcemap:
                azi     = const_or_evo(sourcemap['azimuth'], sstream.sampfracs) * 2 * np.pi
            else:
                azi     = const_or_evo(self.generator.preset['azimuth'], sstream.sampfracs) * 2 * np.pi
            if 'theta' in sourcemap:
                polar   = const_or_evo(sourcemap['theta'], sstream.sampfracs) * np.pi
            elif 'polar' in sourcemap:
                polar   = const_or_evo(sourcemap['polar'], sstream.sampfracs) * np.pi                
            else:
                polar   = const_or_evo(self.generator.preset['polar'], sstream.sampfracs) * np.pi

            # compute sample indices for truncating notes overshooting sonification length
            trunc_note = min(playlen, lastsamp-tsamp)
            trunc_soni   = trunc_note + tsamp

            # spatialise audio by computing relative volume in each speaker
            for i in range(Nchan):
                panenv = self.channels.mics[i].antenna(azi,polar)
                self.out_channels[str(i)].values[tsamp:trunc_soni] += (sstream.values*panenv)[:trunc_note]

        # produce mono audio of caption, if one is provided
        if str(self.caption or '').strip():
            # use a temporary directory to ensure caption file cleanup
            with tempfile.TemporaryDirectory() as cdir:
                cpath = Path(cdir, 'caption.wav')
                render_caption(self.caption, self.samprate,
                               self.ttsmodel, cpath)
                rate_in, wavobj = wavfile.read(cpath)
                wavobj = np.array(wavobj)
            # Set up the Stream objects for TTS
            self.caption_channels = {}
            caption_norm = wavobj.max()
            for c in range(Nchan):
                self.caption_channels[str(c)] = Stream(wavobj.shape[0], self.samprate, ltype='samples')
                
                # place caption straight ahead spatially
                panenv = self.channels.mics[c].antenna(0, 0.5*np.pi)
                
                cnorm = abs(self.out_channels[str(c)].values).max()/caption_norm
                self.caption_channels[str(c)].values += (wavobj*cnorm*panenv)
        else:
            self.caption_channels = {}
            for c in range(Nchan):
                self.caption_channels[str(c)] = Stream(0, self.samprate) 
        
    def save_stereo(self, fname, master_volume=1.):
        """ Save stereo or mono sonifications
        
        Can use this function to save :obj:`"stereo"` or :obj:`"mono"`
        sonifications while avoiding ffmpeg processing.

        Args:
          fname (:obj:`str`) Filename or filepath
          master_volume (:obj:`float`) Amplitude of the largest volume
            peak, from 0-1

        Todo:
          * Support :obj:`master_volume` in decibels
        """

        if len(self.out_channels) > 2:
            print("Warning: sonification has > 2 channels, only first 2 will be used. See 'save_combined' method.")
        
        # first pass - find max amplitude value to normalise output
        # and concatenate channels to list
        vmax = 0.
        channels = []
        for c in range(min(len(self.out_channels), 2)):
            vmax = max(
                abs(self.out_channels[str(c)].values.max()),
                abs(self.out_channels[str(c)].values.min()),
                vmax
            ) / master_volume
            
            # combine caption + sonification streams at display time
            channel_values = np.concatenate([self.out_channels[str(c)].values,
                                self.caption_channels[str(c)].values])   
            
            channels.append(channel_values)
           
        wav.write(fname,
                  np.column_stack(channels),
                  self.samprate, 
                  scale = (-vmax,vmax),
                  sampwidth=3)

        print("Saved.")


    def save_combined(self, fname, ffmpeg_output=False, master_volume=1.):
        """ Save render as a combined multi-channel wav file 
        
        Can use this function to save sonification of any audio_setup,
        using ffmpeg processing, and unscrampling to the correct
        channel order.

        Args:
          fname (:obj:`str`) Filename or filepath
          ffmpeg_output (:obj:`bool`) If True, print :obj:`ffmpeg`
            output to screen 
          master_volume (:obj:`float`) Amplitude of the largest volume
            peak, from 0-1

        Todo:
          * Either find a way to avoid the need to unscramble channle
        	order, or find alternative to save wav files
        """
        # setup list to house wav stream data 
        inputs = [None]*len(self.out_channels)

        # first pass - find max amplitude value to normalise output
        vmax = 0.
        for c in range(len(self.out_channels)):
            vmax = max(
                abs(self.out_channels[str(c)].values.max()),
                abs(self.out_channels[str(c)].values.min()),
                vmax
            ) / master_volume
            
        print("Creating temporary .wav files...")

        # combine caption + sonification streams at display time
        for c in range(len(self.out_channels)):
            tempfname = Path('.', f'.TEMP_{c}.wav')
            self.out_channels[str(c)].values += self.caption_channels[str(c)].values
            wav.write(tempfname, 
                      self.out_channels[str(c)].values,
                      self.samprate, 
                      scale = (-vmax,vmax),
                      sampwidth=3)
            inputs[self.channels.forder[c]] = ff.input(tempfname)
            
        print("Joining temporary .wav files...")
        (
            ff.filter(inputs, 'join', inputs=len(inputs), channel_layout=self.channels.setup)
            .output(fname)
            .overwrite_output()
            .run(quiet=~ffmpeg_output)
        )
        
        print("Cleaning up...")
        for c in range(len(self.out_channels)):
            Path('.', f'.TEMP_{c}.wav').unlink()
            
        print("Saved.")

    def save(self, fname, master_volume=1.):
        """ Save render as a combined multi-channel wav file 
        
        Can use this function to save sonification of any audio_setup
        to a 32-bit depth WAV using `scipy.io.wavfile`

        Args:
          fname (:obj:`str`) Filename or filepath
          master_volume (:obj:`float`) Amplitude of the largest volume
            peak, from 0-1

        Todo:
          * Raise `scipy` issue if common 24-bit WAV can be supported
        """
        
        # first pass - find max amplitude value to normalise output
        vmax = 0.
        for c in range(len(self.out_channels)):
            vmax = max(
                abs(self.out_channels[str(c)].values.max()),
                abs(self.out_channels[str(c)].values.min()),
                vmax
            )

        # normalisation for conversion to int32 bitdepth wav
        norm = master_volume * (pow(2, 31)-1) / vmax

        # setup array to house wav stream data 
        chans = np.zeros((self.out_channels['0'].values.size,
                          len(self.out_channels)), dtype="int32")
        
        # normalise and collect channels into a list
        for c in range(len(self.out_channels)):
            vals = self.out_channels[str(c)].values
            chans[:,c] = (vals*norm).astype("int32")
            
        # finally combine and write out wav file
        wavfile.write(fname, self.samprate, chans)
        print(f"Saved {fname}")

        
    def notebook_display(self, show_waveform=True):
        """ plot the waveforms and embed player in the notebook

        Show waveforms and embed an audio player in the python
        notebook for direct playback. the notebook player only
        supports up to stereo, so if more than two channels, only the
        first two are used as left and right.
        """

        time = self.out_channels['0'].samples / self.out_channels['0'].samprate

        channels = []
        vmax = 0.
        
        # combine caption + sonification streams at display time
        for c in range(len(self.out_channels)):
            channel_values = np.concatenate([self.caption_channels[str(c)].values,
                                             self.out_channels[str(c)].values])   
            channels.append(channel_values)
            vmax = max(
                abs(channels[c].max()),
                abs(channels[c].min()),
                vmax
            ) * 1.05
        
        if show_waveform:
            for i in range(len(self.out_channels)):
                plt.plot(time[::20], self.out_channels[str(i)].values[::20]+2*i*vmax, label=self.channels.labels[i])
            plt.xlabel('Time (s)')
            plt.ylabel('Relative Amplitude')
            plt.legend(frameon=False, loc=5)
            plt.xlim(-time[-1]*0.05,time[-1]*1.2)
            for s in plt.gca().spines.values():
                s.set_visible(False)
                plt.gca().get_yaxis().set_visible(False)
            plt.show()
        
        if len(self.channels.labels) == 1:             
            # we have used 48000 Hz everywhere above as standard, but to quickly hear the sonification sped up / slowed down,
            # you can modify the 'rate' argument below (e.g. multiply by 0.5 for half speed, by 2 for double speed, etc)
            outfmt = np.column_stack(channels*2).T / vmax
        else:
            outfmt = np.column_stack(channels[:2]).T / vmax
        display(ipd.Audio(outfmt,rate=self.out_channels['0'].samprate, autoplay=False))
        
    def hear(self):
        """ Play audio directly to the sound device, for command-line
            playback.
        """

        channels = []
        vmax = 0.
        
        # combine caption + sonification streams at display time
        for c in range(len(self.out_channels)):
            channel_values = np.concatenate([self.caption_channels[str(c)].values,
                                             self.out_channels[str(c)].values])   
            channels.append(channel_values)
            vmax = max(
                abs(channels[c].max()),
                abs(channels[c].min()),
                vmax
            ) * 1.05
                
        if len(self.channels.labels) == 1:             
            # we have used 48000 Hz everywhere above as standard, but to quickly hear the sonification sped up / slowed down,
            # you can modify the 'rate' argument below (e.g. multiply by 0.5 for half speed, by 2 for double speed, etc)
            outfmt = np.column_stack(channels*2)/vmax
        else:
            outfmt = np.column_stack(channels[:2])/vmax

        dur = int(np.round(outfmt.shape[0]/self.out_channels['0'].samprate))
        playback_msg = f"Playing Sonification ({dur} s): "
        print(playback_msg)
        try:
            sd.play(outfmt,self.out_channels['0'].samprate,blocking=1)
        except OSError as error: 
            print(error) 
            print("The Sonification.hear() function requires the PortAudio C-library. This may be missing from your system or \n"
                  "unsupported in this context. This should be installed by pip on Windows and OSx automatically with the \n "
                  "sounddevice library, but on Linux you may need to install manually using e.g.:\n"
                  "\t 'sudo apt-get install libportaudio2.'\n")

    def _make_seamless(self, overlap_dur=0.05):
        self.loop_channels = {}
        buffsize = int(overlap_dur*self.samprate)
        ramp = np.linspace(0,1, buffsize+1)
        for c in range(len(self.out_channels)):
            self.loop_channels[str(c)] = Stream(self.out_channels[str(c)].values.size - buffsize,
                                                self.samprate, ltype='samples')
            self.loop_channels[str(c)].values = self.out_channels[str(c)].values[:-buffsize]
            self.loop_channels[str(c)].values[:buffsize] *= ramp[:-1]
            self.loop_channels[str(c)].values[:buffsize] += ramp[::-1][:-1] * self.out_channels[str(c)].values[-buffsize:]
            



""" The :obj:`score` submodule: musical constraints on what :obj:`Sources` can play

This submodule is intended to control any musical aspects of the
sonification, by providing constraints on harmonic and rhythmic
aspects. 

Todo:
    * Incorporate more rhythmic options, such as quantisation
    * Have the :obj:`Score` class handle pitch and rhythm assignment
      from :obj:`Sources` directly, instead of the :obj:`Sonification`
      class. 
    * Support chord change intervals as a mapped parameter
    * Provide template chord charts for users 
"""

from . import stream
from . import notes
import numpy as np
import pychord as chrd
import numbers
import re

class Score:
    """ Class defining the musical score for the sonification
        
    The Score class controls the musical aspects of the
    sonification. Currently supports a `chordal` score, defining a set
    of notes that can be played for each section of the
    simulation. The score also controls the length of the
    sonification. 

    Note:
    	Currently, no rhythmic constraints are incorporated in the
    	score. Chords are divided evenly over the length of the
    	sonification. For example for a one minute sonification
    	(:obj:`length = '1m 0s'`) the :obj:`chord_sequence = "Am7_3 |
    	D9_3 | Gmaj7_2"` plays each chord for 20s each. Chaining the 
    	same chord can be used to change intervals,
    	(e.g. :obj:`chord_sequence = "F_3 | F_3 | C_4"` plays F for
    	40s and C for 20s.)
 
    Args:
      chord_sequence: (:obj:`str` or :obj:`list`): The chord or chord
    	sequence used for the sonification. If a string, parse using
    	:obj:`parse_chord_sequence`. If a :obj:`list`, each entry is
    	a :obj:`list` of strings or floats, representing the notes of a
        chord. notes are represented as strings using scientific
    	notation, e.g. :obj:`[['C3','E3', 'G3'], ['C3', 'F3', 'A4']]`. If
    	floats, take values as note frequency in Hz. NOTE: currently
    	only supported in compination with the :obj:`Synthesiser`
    	generator class.
      length: (:obj:`str` or :obj:`float`): the length of the
    	sonification. If a string, parse minutes and seconds from
    	format 'Xm Y.Zs'. If a float read as seconds.
      pitch_binning (optional, :obj:`str`): pitch binning mode - choose
        from 'adaptive', where sources are binned by the pitch mapping
        such that each interval is represented the same fraction of the
        time, and 'uniform' where the pitch binning is based on uniform
        size bins in the mapped pitch parameter. 
    """
    def __init__(self, chord_sequence, length, pitch_binning='adaptive'):
        # check types to handle score length correctly
        if isinstance(length, str):
            regex = "([0-9]*)m\s*([0-9]*.[0-9]*)s"
            reobj = re.match(regex, length, re.M | re.I)
            self.length = float(reobj.group(1))*60. + float(reobj.group(2))            
        else:
            self.length = length        
        
        # check types to handle different chord formats correctly
        if isinstance(chord_sequence, list):
            self.note_sequence = chord_sequence
        if isinstance(chord_sequence, str):
            self.note_sequence = parse_chord_sequence(chord_sequence)

        if pitch_binning in ['adaptive','uniform']:
            self.pitch_binning = pitch_binning
        else:
            raise Exception(
                f"\"{pitch_binning}\" is not a valid pitch_binning mode")        

        # number of chords in the sequence 
        self.nchords = len(self.note_sequence)
        self.nintervals = [len(c) for c in self.note_sequence]
        
        # For now, chords changes are just equally spaced in the timeline
        self.fracbins = np.linspace(0,1, self.nchords+1) 
        self.timebins = self.length * self.fracbins
        
def parse_chord_sequence(chord_sequence):
    """ parse a chord sequence from a string

    Args:
      chord_sequence (:obj:`str`): chord sequence to parse, with chord
    	names appended with '_N' where N is the root octave of the
    	chord, and each chord is separated by a pipe character, '|'.

    Returns:
      note_list (:obj:`list(list)`): the chord sequence represented as
    	a list of lists, where each sub-list is a chord comprised of
    	strings representing each note in scentific notation (e.g. 'A4') 
    """
    chord_list = chord_sequence.split("|")
    note_list = []
    note_frqs = []
    for i in range(len(chord_list)):
        chord_list[i] = chord_list[i].strip()
        chord_list[i] = chord_list[i].split('_')
        chord_notes = notes.chord_notes(*chord_list[i])
        note_list.append(chord_notes)
    return note_list

if __name__ == '__main__':
    chords = "Am7_4 | G_3 | F_3 | E7b9_3 "
    # chords = "Am7_4"
    Score(chords)



from scipy.io import wavfile
from scipy.interpolate import interp1d
import numpy as np
import strauss.utilities as utils
import re
try:
    from TTS.api import TTS
except (OSError, ModuleNotFoundError) as sderr:
    def TTS(*args, **kwargs):
        raise TTSIsNotSupported("strauss has not been installed with text-to-speech support. \n"
              "This is not installed by default, due to some specific module requirements of the TTS module."
              "Reinstalling strauss with 'pip install strauss[TTS]' will give you access to this function")

class TTSIsNotSupported(Exception):
    pass

def render_caption(caption, samprate, model, caption_path):
    '''The render_caption function generates an audio caption from text input
    and writes it as a wav file. If the sample rate of the model is not equal 
    to that passed from sonification.py, it resamples to the correct rate and
    re-writes the file. Text from user input is converted with text-to-speech
    software from Coqui-AI - https://pypi.org/project/TTS/ . You can view 
    publicly available voice models with 'TTS.list_models()'

    Args:
      caption (:obj:`str`): script to be spoken by the TTS voice
      samprate (:obj:`int`): samples per second
      model (:obj:`str`): valid name of TTS voice from the underying TTS
        module
      model (:obj:`str`): valid name of TTS voice from the underying TTS
        module
      caption_path (:obj:`str`): filepath for spoken caption output
    '''

    # TODO: do this better with logging. We can filter TTS function output, e.g. alert to downloading models...
    print('Rendering caption (this can take a while if the caption is long, or if the TTS model needs downloading)...')
    
    # capture stdout from the talkative TTS module
    with utils.Capturing() as output:
        # Load in the tts model
        tts = TTS(model, progress_bar=False, gpu=False)

        # render to speech, and write as a wav file (allow )
        tts.tts_to_file(text=caption, file_path=caption_path)

        
    # Read the file back in to check the sample rate
    rate_in, wavobj = wavfile.read(caption_path)
    
    #If it doesn't match the required rate, resample and re-write
    if rate_in != samprate:
        new_wavobj = utils.resample(rate_in, samprate, wavobj)
        wavfile.write(caption_path, samprate, new_wavobj)





""" The :obj:`sources` submodule: representing data as sound sources.

This submodule deals with the mapping of input datasets to the parameters
controlling sound in the eventual sonification.   

Attributes:
   mappable (:obj:`list(str)`): List of strings indicating possible
	sonification parameters to which data can be mapped.
   evolvable (:obj:`list(str)`): List of strings indicating the subset of
	`mappable` parameters that can be evolved continuosly for an
	individual Source.
   param_limits (:obj:`list(tuple)`): List of tuples indicating the default
	numerical ranges bounding corresponding mappable parameter
	(e.g. 0-1 for volume).
   param_lim_dict (:obj:`dict`): Dictionary combining `mappable` (keys) and 
	`param_limits` (items).

Todo:
    * Store mappable, evolvable and parameter ranges in YAML files (cleaner). 
    * Specialised Event and Object child classes (eg. spectralisation).
"""

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
from .utilities import rescale_values 

mappable = ['polar',
            'azimuth',
            'theta',
            'phi',
            'volume',
            'pitch',
            'time',
            'cutoff',
            'time_evo',
            'spectrum',
            'pitch_shift',
            'volume_envelope/A',
            'volume_envelope/D',
            'volume_envelope/S',
            'volume_envelope/R',
            'volume_lfo/freq',
            'volume_lfo/freq_shift',
            'volume_lfo/amount',
            'pitch_lfo/freq',
            'pitch_lfo/freq_shift',
            'pitch_lfo/amount']

evolvable = ['polar',
             'azimuth',
             'theta',
             'phi',
             'volume',
             'cutoff',
             'time_evo',
             'pitch_shift',
             'volume_lfo/freq',
             'volume_lfo/freq_shift',
             'volume_lfo/amount',
             'pitch_lfo/freq',
             'pitch_lfo/freq_shift',
             'pitch_lfo/amount']
param_limits = [(0,1),#np.pi),
                (0,1),#2*np.pi),
                (0,1),#np.pi),
                (0,1),#2*np.pi),
                (0,1),
                (0,1),
                (0,1),
                (0,1),
                (0,1),
                (0,1),
                (0,24),
                (1e-2, 10),
                (1e-2, 10),
                (0,1),
                (1e-2, 10),
                (1,12),
                (0,3),
                (0,2),
                (1,12),
                (0,3),
                (0,1)]

param_lim_dict = dict(zip(mappable, param_limits))

class Source:
    """ Generic source class defining common methods/attributes
    
    `Source` and its child classes represent the input data, and its
    mapping to sonification parameters.

    Note:
	`Source` isn't used directly, instead use child classes
    	`Events` or `Objects`.

    Args:
    	mapped_quantities (:obj:`list(str)`): The subset of parameters to
    	   which data will be mapped. 

    Raises:
    	UnrecognisedProperty: if `mapped_quantities` entry not in `mappable`. 

    """
    def __init__(self, mapped_quantities):
        # check these are all mappable parameters

        
        for q in mapped_quantities:
            if q not in mappable:
                raise UnrecognisedProperty(
                    f"Property \"{q}\" is not recognised")

        if ('theta' in mapped_quantities) and ('polar' in mapped_quantities):
            raise Exception(
                "\"theta\" and \"polar\" cannot be combined as " \
                "these represent the same quantity: \"theta\" and " \
                "\"phi\" are deprecated and will be replaced with \"polar\"" \
                " and \"azimuth\" in a future version.")

        if ('phi' in mapped_quantities) and ('azimuth' in mapped_quantities):
            raise Exception(
                "\"phi\" and \"azimuth\" cannot be combined as " \
                "these represent the same quantity: \"theta\" and " \
                "\"phi\" are deprecated and will be replaced with \"polar\"" \
                " and \"azimuth\" in a future version.")            
            
        # initialise common structures
        self.mapped_quantities = mapped_quantities
        self.raw_mapping = {}
        self.mapping = {}
        self.mapping_evo = {}
        
    def apply_mapping_functions(self, map_funcs={}, map_lims={}, param_lims={}):
        """ Taking input data and mapping to parameters.

        This function does the bulk of the work for `Source` classes,
        taking each input data variable and applying the mapping
        function (x' = x by default), descaling by the x' upper and
        lower limits and rescaling to the sonification parameter
        limits. These values are stored for non-evolving parameters,
        while for evolving properties are converted to interpolation
        functions. 

        Args:
    	   map_funcs (:obj:`dict`, optional): dict with keys that must be
        	a subset self.mapped_quantities. Entries are then
        	function-like objects for converting input data
        	(e.g. taking log of a data set). If not provided,
        	each conversion function is assumed to be  f(x) = x.  
           map_lims (:obj:`dict`, optional): dict with keys that must be
        	a subset self.mapped_quantities. Entries are
        	tuples indicating the lower (index 0) and upper (index
        	1) limits on the converted input data
        	values. numerical values indicate absolute limits,
        	while strings indicate percentiles
        	[e.g. ('10','95')]. converted data values are clipped
        	to these limits. If not provided, (0,1) is assumed.
           param_lims (:obj:`dict`, optional): dict with keys that
        	must be a subset self.mapped_quantities. Entries are
        	tuples indicating the lower (index 0) and upper (index
        	1) limits of the mapped sonification parameters. The
        	map_lims ranges are resaled to these ranges to give
        	the parameter values. If not provided, the default
        	param_lim_dict values are taken.

        Note:
           There is special behaviour for the `polar` and `azimuth`
           parameters, to ensure shortest angular distance when
           interpolating across the 0-2pi and 0-pi boundaries.
        
        """
        for key in self.mapped_quantities:
            rawvals = self.raw_mapping[key]

            # apply mapping functions if specified
            if key in map_funcs:
                mapvals = map_funcs[key](rawvals)
            else:
                mapvals = rawvals

            # set mapping limits if specified
            if key in map_lims:
                vallims = map_lims[key]
            else:
                vallims = (0,1)

            # set parameter limits if specified
            if key in param_lims:
                plims = param_lims[key]
            else:
                plims = param_lim_dict[key]
                
            lims = []
            # scale mapped values within limits if specified
            for l in vallims:
                if isinstance(l, str):
                    # string values notate percentile limits
                    pc = float(l)
                    buff = 1
                    sub = 0
                    if pc > 100:
                        # if percentile over 100 we add 
                        buff = pc/100.
                        pc = 100
                        sub = lims[0]
                    lim = sub + (np.percentile(np.hstack([mapvals]), pc) - sub)*buff
                    lims.append(lim)
                else:
                    # numerical values notate absolute limits
                    lims.append(l)
    
            # limit mapped values from 0 to 1 NOTE: do we want to mix and match const and evo?
            if hasattr(mapvals[0], "__iter__"):
                self.mapping[key] = []
                for i in range(self.n_sources):
                    scaledvals = rescale_values(mapvals[i], lims, plims)
                    self.mapping[key].append(scaledvals)
            else:
                scaledvals = rescale_values(np.array(mapvals), lims, plims)
                self.mapping[key] =  list(scaledvals)
            
        # finally, iterate through sources and interpolate evo functions 
        for key in self.mapping:
            if key == "time_evo":
                continue
            if key == "spectrum":
                # if hasattr(self.mapping[key][0][0], "__iter__"):
                # ^ in case we want to catch and pre process multi-spectra
                continue
            elif hasattr(self.mapping[key][0], "__iter__"):
                # print(key, self.mapping[key][0])
                for i in range(self.n_sources):
                    if key not in evolvable:
                        raise Exception(f"Mapping error: Parameter \"{key}\" cannot be evolved.")
                    x = self.mapping["time_evo"][i]
                    y = self.mapping[key][i]
                    if key == "phi" or key == "azimuth":
                        # special case: shortest angular distance
                        # between phi points is always assumed
                        ydiff = np.diff(y)
                        discont_bdx = abs(ydiff) > 0.5
                        for j in range(discont_bdx.sum()):
                            xpre = x[:-1][discont_bdx][j]
                            ysense = np.sign(ydiff[discont_bdx][j]) 
                            y[x > xpre] -= ysense
                    self.mapping[key][i] = interp1d(x,y, bounds_error=False,
                                                    fill_value=(y[0],y[-1]))
            
class Events(Source):
    """ Represent data as time-discrete events.

    Child class of `Source`, for `Event`-type sources. Each `Event` is
    discrete in `time` with single data values mapped to each
    sonification parameter. 
    """
    def fromfile(self, datafile, coldict):
        """Take input data from ASCII file

        Args:
          datafile (:obj:`str`): path to input data file
          coldict (:obj:`dict`): keys are self.mapped_values, with
        	entries integer indexes for their corresponding column.
        """
        data = np.genfromtxt(datafile)
        for key in self.mapped_quantities:
            self.raw_mapping[key] = data[:,coldict[key]] 
        self.n_sources = data.shape[0]
        
    def fromdict(self, datadict):
        """Take input data from dictionary
	
        Args:
          datadict (:obj:`dict`): keys are self.mapped_values, with
        	entries corresponding to the input data. Multiple
        	sources are provided as :obj:`lists`, with data for
       		each source corresponding to the values. Single
        	sources can be represented as single values.
        """
        for key in self.mapped_quantities:
            if key in datadict:
                self.raw_mapping[key] = datadict[key]
            else:
                Exception(f"Mapped property {key} not in datadict.")
        self.n_sources = np.array(datadict[key]).shape[0]
 
class Objects(Source):
    """ Represent data as time-continuous objects.
    
    Child class of `Source`. In addition to supporting single values
    for each parameter (see `Events` class), objects also support
    time evolution for `evolvable` parameters, given a `time-evo`
    mapping.

    Todo:
    	* implement :obj:`fromfile` method
    """
    def fromdict(self, datadict):
        """ Take input data from dictionary
	
        Args:
          datadict (:obj:`dict`): keys are self.mapped_values, with
        	entries corresponding to the input data. Multiple
        	sources are provided as either :obj:`lists` or 2D
        	:obj:`numpy.array` objects, with each source
        	corresponding to the entries or columns respectively.
        	Single sources can be represented as single values or
        	1D :obj:`numpy.array` (for evolving parameters). 
        """
        for key in self.mapped_quantities:
            if key in datadict:
                d = datadict[key]
                if (type(d) is not list) and (np.array(d).ndim <= 1):
                    self.raw_mapping[key] = [d]
                else:
                    self.raw_mapping[key] = d
            else:
                Exception(f"Mapped property {key} not in datadict.")
        self.n_sources = np.array(self.raw_mapping[key]).shape[0]

class UnrecognisedProperty(Exception):
    "Error raised when trying to map unrecognised parameters"
    pass
